# Copyright 2015 Ocado Innovation Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Models for closuretree application."""

# We like magic.
# pylint: disable=W0142

# We have lots of dynamically generated things, hard for pylint to solve.
# pylint: disable=E1101

# It may not be our class, but we made the attribute on it
# pylint: disable=W0212

# Public methods are useful!
# pylint: disable=R0904

from django.db import models
from django.db.models.base import ModelBase
from django.db.models.signals import post_save, pre_delete
from django.dispatch import receiver
from django.utils.six import with_metaclass
import sys

def _closure_model_unicode(self):
    """__unicode__ implementation for the dynamically created
        <Model>Closure model.
    """
    return "Closure from %s to %s" % (self.parent, self.child)

def create_closure_model(cls):
    """Creates a <Model>Closure model in the same module as the model."""
    meta_vals = {
        'unique_together':  (("parent", "child"),)
    }
    if getattr(cls._meta, 'db_table', None):
        meta_vals['db_table'] = '%sclosure' % getattr(cls._meta, 'db_table')
    model = type('%sClosure' % cls.__name__, (models.Model,), {
        'parent': models.ForeignKey(
            cls.__name__,
            related_name=cls.closure_parentref()
        ),
        'child': models.ForeignKey(
            cls.__name__,
            related_name=cls.closure_childref()
        ),
        'depth': models.IntegerField(),
        '__module__':   cls.__module__,
        '__unicode__': _closure_model_unicode,
        'Meta': type('Meta', (object,), meta_vals),
    })
    setattr(cls, "_closure_model", model)
    return model

class ClosureModelBase(ModelBase):
    """Metaclass for Models inheriting from ClosureModel,
        to ensure the <Model>Closure model is created.
    """
    #This is a metaclass. MAGIC!
    def __init__(cls, name, bases, dct):
        """Create the closure model in addition
            to doing all the normal django stuff.
        """
        super(ClosureModelBase, cls).__init__(name, bases, dct)
        if not cls._meta.get_parent_list() and cls.__module__ != __name__:
            setattr(
                sys.modules[cls.__module__],
                '%sClosure' % cls.__name__,
                create_closure_model(cls)
            )

class ClosureModel(with_metaclass(ClosureModelBase, models.Model)):
    """Provides methods to assist in a tree based structure."""
    # pylint: disable=W5101

    class Meta:
        """We make this an abstract class, it needs to be inherited from."""
        # pylint: disable=W0232
        # pylint: disable=R0903
        abstract = True

    def __setattr__(self, name, value):
        if name.endswith('_id'):
            id_field_name = name
        else:
            id_field_name = "%s_id" % name
        if (
            name.startswith(self._closure_sentinel_attr) and  # It's the right attribute
            (  # It's already been set
                (hasattr(self, 'get_deferred_fields') and id_field_name not in self.get_deferred_fields() and hasattr(self, id_field_name)) or  # Django>=1.8
                (not hasattr(self, 'get_deferred_fields') and hasattr(self, id_field_name))  # Django<1.8
            ) and
            not self._closure_change_check()  # The old value isn't stored
        ):
            if name.endswith('_id'):
                obj_id = value
            elif value:
                obj_id = value.pk
            else:
                obj_id = None
            # If this is just setting the same value again, we don't need to do anything
            if getattr(self, id_field_name) != obj_id:
                # Already set once, and not already stored the old
                # value, need to take a copy before it changes
                self._closure_change_init()
        super(ClosureModel, self).__setattr__(name, value)

    @classmethod
    def _toplevel(cls):
        """Find the top level of the chain we're in.

            For example, if we have:
            C inheriting from B inheriting from A inheriting from ClosureModel
            C._toplevel() will return A.
        """
        superclasses = (
            list(set(ClosureModel.__subclasses__()) &
                 set(cls._meta.get_parent_list()))
        )
        return next(iter(superclasses)) if superclasses else cls

    @classmethod
    def rebuildtable(cls):
        """Regenerate the entire closuretree."""
        cls._closure_model.objects.all().delete()
        cls._closure_model.objects.bulk_create([cls._closure_model(
            parent_id=x['pk'],
            child_id=x['pk'],
            depth=0
        ) for x in cls.objects.values("pk")])
        for node in cls.objects.all():
            node._closure_createlink()

    @classmethod
    def closure_parentref(cls):
        """How to refer to parents in the closure tree"""
        return "%sclosure_children" % cls._toplevel().__name__.lower()

    # Backwards compatibility:
    _closure_parentref = closure_parentref

    @classmethod
    def closure_childref(cls):
        """How to refer to children in the closure tree"""
        return "%sclosure_parents" % cls._toplevel().__name__.lower()

    # Backwards compatibility:
    _closure_childref = closure_childref

    @property
    def _closure_sentinel_attr(self):
        """The attribute we need to watch to tell if the
            parent/child relationships have changed
        """
        meta = getattr(self, 'ClosureMeta', None)
        return getattr(meta, 'sentinel_attr', self._closure_parent_attr)

    @property
    def _closure_parent_attr(self):
        '''The attribute or property that holds the parent object.'''
        meta = getattr(self, 'ClosureMeta', None)
        return getattr(meta, 'parent_attr', 'parent')

    @property
    def _closure_parent_pk(self):
        """What our parent pk is in the closure tree."""
        if hasattr(self, "%s_id" % self._closure_parent_attr):
            return getattr(self, "%s_id" % self._closure_parent_attr)
        else:
            parent = getattr(self, self._closure_parent_attr)
            return parent.pk if parent else None

    def _closure_deletelink(self, oldparentpk):
        """Remove incorrect links from the closure tree."""
        self._closure_model.objects.filter(
            **{
                "parent__%s__child" % self._closure_parentref(): oldparentpk,
                "child__%s__parent" % self._closure_childref(): self.pk
            }
        ).delete()

    def _closure_createlink(self):
        """Create a link in the closure tree."""
        linkparents = self._closure_model.objects.filter(
            child__pk=self._closure_parent_pk
        ).values("parent", "depth")
        linkchildren = self._closure_model.objects.filter(
            parent__pk=self.pk
        ).values("child", "depth")
        newlinks = [self._closure_model(
            parent_id=p['parent'],
            child_id=c['child'],
            depth=p['depth']+c['depth']+1
        ) for p in linkparents for c in linkchildren]
        self._closure_model.objects.bulk_create(newlinks)

    def get_ancestors(self, include_self=False, depth=None):
        """Return all the ancestors of this object."""
        if self.is_root_node():
            if not include_self:
                return self._toplevel().objects.none()
            else:
                # Filter on pk for efficiency.
                return self._toplevel().objects.filter(pk=self.pk)

        params = {"%s__child" % self._closure_parentref():self.pk}
        if depth is not None:
            params["%s__depth__lte" % self._closure_parentref()] = depth
        ancestors = self._toplevel().objects.filter(**params)
        if not include_self:
            ancestors = ancestors.exclude(pk=self.pk)
        return ancestors.order_by("%s__depth" % self._closure_parentref())

    def get_descendants(self, include_self=False, depth=None):
        """Return all the descendants of this object."""
        params = {"%s__parent" % self._closure_childref():self.pk}
        if depth is not None:
            params["%s__depth__lte" % self._closure_childref()] = depth
        descendants = self._toplevel().objects.filter(**params)
        if not include_self:
            descendants = descendants.exclude(pk=self.pk)
        return descendants.order_by("%s__depth" % self._closure_childref())

    def prepopulate(self, queryset):
        """Perpopulate a descendants query's children efficiently.
            Call like: blah.prepopulate(blah.get_descendants().select_related(stuff))
        """
        objs = list(queryset)
        hashobjs = dict([(x.pk, x) for x in objs] + [(self.pk, self)])
        for descendant in hashobjs.values():
            descendant._cached_children = []
        for descendant in objs:
            assert descendant._closure_parent_pk in hashobjs
            parent = hashobjs[descendant._closure_parent_pk]
            parent._cached_children.append(descendant)

    def get_children(self):
        """Return all the children of this object."""
        if hasattr(self, '_cached_children'):
            children = self._toplevel().objects.filter(
                pk__in=[n.pk for n in self._cached_children]
            )
            children._result_cache = self._cached_children
            return children
        else:
            return self.get_descendants(include_self=False, depth=1)

    def get_root(self):
        """Return the furthest ancestor of this node."""
        if self.is_root_node():
            return self

        return self.get_ancestors().order_by(
            "-%s__depth" % self._closure_parentref()
        )[0]

    def is_child_node(self):
        """Is this node a child, i.e. has a parent?"""
        return not self.is_root_node()

    def is_root_node(self):
        """Is this node a root, i.e. has no parent?"""
        return self._closure_parent_pk is None

    def is_descendant_of(self, other, include_self=False):
        """Is this node a descendant of `other`?"""
        if other.pk == self.pk:
            return include_self

        return self._closure_model.objects.filter(
            parent=other,
            child=self
        ).exclude(pk=self.pk).exists()

    def is_ancestor_of(self, other, include_self=False):
        """Is this node an ancestor of `other`?"""
        return other.is_descendant_of(self, include_self=include_self)

    def _closure_change_init(self):
        """Part of the change detection. Setting up"""
        # More magic. We're setting this inside setattr...
        # pylint: disable=W0201
        self._closure_old_parent_pk = self._closure_parent_pk

    def _closure_change_check(self):
        """Part of the change detection. Have we changed since we began?"""
        return hasattr(self,"_closure_old_parent_pk")

    def _closure_change_oldparent(self):
        """Part of the change detection. What we used to be"""
        return self._closure_old_parent_pk


@receiver(post_save, dispatch_uid='closure-model-save')
def closure_model_save(sender, **kwargs):
    if issubclass(sender, ClosureModel):
        instance = kwargs['instance']
        create = kwargs['created']
        if create:
            closure_instance = instance._closure_model(
                parent=instance,
                child=instance,
                depth=0
            )
            closure_instance.save()
        if instance._closure_change_check():
            #Changed parents.
            if instance._closure_change_oldparent():
                instance._closure_deletelink(instance._closure_change_oldparent())
            instance._closure_createlink()
            delattr(instance, "_closure_old_parent_pk")
        elif create:
            # We still need to create links when we're first made
            instance._closure_createlink()


@receiver(pre_delete, dispatch_uid='closure-model-delete')
def closure_model_delete(sender, **kwargs):
    if issubclass(sender, ClosureModel):
        instance = kwargs['instance']
        instance._closure_deletelink(instance._closure_parent_pk)



# Copyright 2015 Ocado Innovation Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Setup file for django-closuretree."""
from setuptools import setup, find_packages

from closuretree.version import __VERSION__

setup(
    name='django-closuretree',
    version=__VERSION__,
    packages=find_packages(),
    author='Mike Bryant',
    author_email='mike.bryant@ocado.com',
    description='Efficient tree-based datastructure for Django',
    long_description=open('README.rst').read(),
    url='https://github.com/ocadotechnology/django-closuretree/',
    install_requires=[
        'django >= 1.4, < 1.12',
        'django-autoconfig',
    ],
    tests_require=['django-setuptest >= 0.2'],
    test_suite='setuptest.setuptest.SetupTestSuite',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Framework :: Django',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.2',
        'Programming Language :: Python :: 3.3',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
)



# Copyright 2015 Ocado Innovation Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for closuretree"""

# This is kinda tricky...
# pylint: disable=C0103
# pylint: disable=E1101
# pylint: disable=E0602
# pylint: disable=R0903
# pylint: disable=R0904
# pylint: disable=
# pylint: disable=

from django import VERSION
from django.test import TestCase
from django.db import models
from closuretree.models import ClosureModel
import uuid

class TC(ClosureModel):
    """A test model."""
    parent2 = models.ForeignKey(
        "self",
        related_name="children",
        null=True,
        blank=True
    )
    name = models.CharField(max_length=32)
    blah = models.ForeignKey("Blah", related_name="tcs", null=True, blank=True)

    class ClosureMeta(object):
        """Closure options."""
        parent_attr = "parent2"

    def __unicode__(self):
        return "%s: %s" % (self.pk, self.name)

class Blah(models.Model):
    """A test model for foreign keys"""
    thing = models.CharField(max_length=32)

class TCSUB(TC):
    """Testing closure subclasses."""
    extrafield = models.IntegerField()

class TCSUB2(TCSUB):
    """Testing closure subclasses."""
    ef = models.IntegerField()

class BaseTestCase(TestCase):
    """Providing details for testing."""

    normal_model = TC
    closure_model = TCClosure

    def setUp(self):
        self.a = self.normal_model.objects.create(name="a")
        self.b = self.normal_model.objects.create(name="b")
        self.c = self.normal_model.objects.create(name="c")
        self.d = self.normal_model.objects.create(name="d")

        # We cheat here, we don't care about the __unicode__ method,
        # It's only useful when we're working out why the tests fail.
        self.a.__unicode__()

    def test_unicode(self):
        """Test unicoding of the closures works."""
        # No, it's a method for unittest!
        # pylint: disable=R0201

        for obj in self.closure_model.objects.all():
            obj.__unicode__()

    def test_adding(self):
        """
        Tests that adding a new parent relationship creates closures
        """
        self.failUnlessEqual(self.closure_model.objects.count(), 4)
        self.b.parent2 = self.a
        self.b.save()
        self.failUnlessEqual(self.closure_model.objects.count(), 5)
        self.c.parent2 = self.b
        self.c.save()
        # Test double save
        self.c.save()
        self.d.parent2 = self.c
        self.d.save()
        self.failUnlessEqual(self.closure_model.objects.count(), 10)

    def test_deletion(self):
        """
            Tests that deleting a relationship removes the closure entries.
        """
        self.failUnlessEqual(self.closure_model.objects.count(), 4)
        self.b.parent2 = self.a
        self.b.save()
        self.failUnlessEqual(self.closure_model.objects.count(), 5)
        self.b.parent2 = None
        self.b.save()
        self.failUnlessEqual(self.closure_model.objects.count(), 4)
        self.b.parent2 = self.a
        self.b.save()
        self.c.parent2 = self.b
        self.c.save()
        self.failUnlessEqual(self.closure_model.objects.count(), 7)
        self.b.delete()
        self.failUnlessEqual(self.closure_model.objects.count(), 2)


if VERSION >= (1, 8):
    class UUIDTC(ClosureModel):
        '''Testing support for models with a UUID primary key. See #30'''
        primary_key = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
        parent2 = models.ForeignKey(
            "self",
            related_name="children",
            null=True,
            blank=True,
        )
        name = models.CharField(max_length=32)

        class ClosureMeta(object):
            """Closure options."""
            parent_attr = "parent2"

        def __unicode__(self):
            return "%s: %s" % (self.pk, self.name)


    class UUIDTestCase(BaseTestCase):
        '''Testing UUID models. See #30'''
        normal_model = UUIDTC
        closure_model = UUIDTCClosure


class AncestorTestCase(TestCase):
    """Testing things to do with ancestors."""

    def setUp(self):
        self.a = TC.objects.create(name="a")
        self.b = TC.objects.create(name="b")
        self.c = TC.objects.create(name="c")
        self.b.parent2 = self.a
        self.b.save()
        self.c.parent2 = self.b
        self.c.save()

    def test_ancestors(self):
        """Testing the ancestors method."""
        self.failUnlessEqual(list(self.a.get_ancestors()), [])
        self.failUnlessEqual(list(self.b.get_ancestors()), [self.a])
        self.failUnlessEqual(
            list(self.a.get_ancestors(include_self=True)),
            [self.a]
        )
        self.failUnlessEqual(
            list(self.c.get_ancestors(include_self=True)),
            [self.c, self.b, self.a]
        )
        self.failUnlessEqual(
            list(self.c.get_ancestors(include_self=True, depth=1)),
            [self.c, self.b]
        )

    def test_descendants(self):
        """Testing the descendants method."""
        self.failUnlessEqual(list(self.c.get_descendants()), [])
        self.failUnlessEqual(list(self.b.get_descendants()), [self.c])
        self.failUnlessEqual(
            list(self.a.get_descendants(include_self=True)),
            [self.a, self.b, self.c]
        )
        self.failUnlessEqual(
            list(self.c.get_descendants(include_self=True)),
            [self.c]
        )

    def test_children(self):
        """Testing the children method."""
        self.failUnlessEqual(list(self.c.get_children()), [])
        self.failUnlessEqual(list(self.b.get_children()), [self.c])

class RebuildTestCase(TestCase):
    """Test rebuilding the tree"""

    def setUp(self):
        self.a = TC.objects.create(name="a")
        self.b = TC.objects.create(name="b")
        self.b.parent2 = self.a
        self.b.save()
        self.c = TC.objects.create(name="c")
        self.c.parent2 = self.b
        self.c.save()
        self.d = TC.objects.create(name="d")
        self.d.parent2 = self.a
        self.d.save()

    def test_rebuild_from_full(self):
        """Test a rebuild when the tree is correct."""

        self.failUnlessEqual(TCClosure.objects.count(), 8)
        TC.rebuildtable()
        self.failUnlessEqual(TCClosure.objects.count(), 8)

    def test_rebuild_from_empty(self):
        """Test a rebuild when the tree is empty."""

        TCClosure.objects.all().delete()
        TC.rebuildtable()
        self.failUnlessEqual(TCClosure.objects.count(), 8)

    def test_rebuild_from_partial(self):
        """Test a rebuild when the tree is partially empty."""

        TCClosure.objects.get(parent__name='a', child__name='a').delete()
        TCClosure.objects.get(parent__name='a', child__name='c').delete()
        self.failUnlessEqual(TCClosure.objects.count(), 6)
        TC.rebuildtable()
        self.failUnlessEqual(TCClosure.objects.count(), 8)

class InitialClosureTestCase(TestCase):
    """Tests for when things are created with a parent."""

    def test_creating_with_parent(self):
        """Make sure closures are created when making objects."""
        a = TC.objects.create(name="a")
        self.failUnlessEqual(TCClosure.objects.count(), 1)
        b = TC.objects.create(name="b", parent2=a)
        self.failUnlessEqual(TCClosure.objects.count(), 3)
        TC.objects.create(name="c", parent2=b)
        self.failUnlessEqual(TCClosure.objects.count(), 6)

class IsTestCase(TestCase):
    """Test some useful methods."""

    def setUp(self):
        self.a = TC.objects.create(name="a")
        self.b = TC.objects.create(name="b", parent2=self.a)
        self.c = TC.objects.create(name="c", parent2=self.b)
        self.d = TC.objects.create(name="d", parent2=self.c)
        self.e = TC.objects.create(name="e", parent2=self.b)
        self.f = TC.objects.create(name="f", parent2=self.e)

    def test_ancestor_of(self):
        """Test is_ancestor_of method."""
        self.assertEqual(self.a.is_ancestor_of(self.c), True)
        self.assertEqual(self.a.is_ancestor_of(self.d), True)
        self.assertEqual(self.a.is_ancestor_of(self.e), True)
        self.assertEqual(self.d.is_ancestor_of(self.b), False)
        self.assertEqual(
            self.d.is_ancestor_of(self.b, include_self=True),
            False
        )
        self.assertEqual(self.f.is_ancestor_of(self.f), False)
        self.assertEqual(
            self.f.is_ancestor_of(self.f, include_self=False),
            False
        )
        self.assertEqual(self.f.is_ancestor_of(self.f, include_self=True), True)

    def test_descendant_of(self):
        """Test id_descendant_of method."""
        self.assertEqual(self.a.is_descendant_of(self.a), False)
        self.assertEqual(self.f.is_descendant_of(self.f), False)
        self.assertEqual(
            self.a.is_descendant_of(self.a, include_self=False), False
        )
        self.assertEqual(
            self.a.is_descendant_of(self.a, include_self=True), True
        )
        self.assertEqual(self.a.is_descendant_of(self.c), False)
        self.assertEqual(self.c.is_descendant_of(self.a), True)

    def test_get_root(self):
        """Test get_root method"""
        self.assertEqual(self.a.get_root(), self.a)
        self.assertEqual(self.b.get_root(), self.a)
        self.assertEqual(self.f.get_root(), self.a)

    def test_child_node(self):
        """Test is_child_node method"""
        self.assertEqual(self.a.is_child_node(), False)
        self.assertEqual(self.b.is_child_node(), True)
        self.assertEqual(self.f.is_child_node(), True)

class PrepopulateTestCase(TestCase):
    """Test prepopulating."""

    def setUp(self):
        self.a = TC.objects.create(name="a")
        self.b = TC.objects.create(name="b", parent2=self.a)
        self.c = TC.objects.create(name="c", parent2=self.b)
        self.d = TC.objects.create(name="d", parent2=self.c)
        self.e = TC.objects.create(name="e", parent2=self.b)
        self.f = TC.objects.create(name="f", parent2=self.e)

    def test_prepopulate(self):
        """Test prepopulating"""
        with self.assertNumQueries(6):
            children = []
            for node in self.a.get_descendants():
                children.extend(list(node.get_children()))
            self.assertEqual(len(children), 4)
        with self.assertNumQueries(1):
            children = []
            queryset = self.a.get_descendants()
            self.a.prepopulate(queryset)
            for node in queryset:
                children.extend(list(node.get_children()))
            self.assertEqual(len(children), 4)

    def test_prepopulate_not_root(self):
        """Test prepopulating when we're not the root"""
        with self.assertNumQueries(5):
            children = []
            for node in self.b.get_descendants():
                children.extend(list(node.get_children()))
            self.assertEqual(len(children), 2)
        with self.assertNumQueries(1):
            children = []
            queryset = self.b.get_descendants()
            self.b.prepopulate(queryset)
            for node in queryset:
                children.extend(list(node.get_children()))
            self.assertEqual(len(children), 2)

class SentinelModel(ClosureModel):
    """A model using a sentinel attribute."""
    location = models.ForeignKey(
        "IntermediateModel",
        null=True,
        blank=True
    )

    @property
    def parent(self):
        """Return the object's parent."""
        if self.location:
            return self.location.real_parent

    class ClosureMeta(object):
        """Closure options."""
        sentinel_attr = "location"
        parent_attr = "parent"

class IntermediateModel(models.Model):
    """The intermediate model between the sentinel model
        and its parent (itself).
    """
    real_parent = models.ForeignKey(
        'SentinelModel',
        null=True,
        blank=True,
    )

class SentinelAttributeTestCase(TestCase):
    """Test functionality of the sentinel attribute."""

    def setUp(self):
        self.a = SentinelModel.objects.create()
        self.b = SentinelModel.objects.create()
        self.c = SentinelModel.objects.create()
        self.d = SentinelModel.objects.create()
        self.l1 = IntermediateModel.objects.create(real_parent=self.a)
        self.l2 = IntermediateModel.objects.create(real_parent=self.b)
        self.l3 = IntermediateModel.objects.create(real_parent=self.c)

    def test_closure_creation(self):
        '''Test creation of closures in the sentinel case'''

        self.failUnlessEqual(SentinelModelClosure.objects.count(), 4)

        self.b.location = self.l1
        self.b.save()
        self.failUnlessEqual(self.b.parent, self.a)
        self.c.location = self.l2
        self.c.save()

        self.failUnlessEqual(SentinelModelClosure.objects.count(), 7)

    def test_num_queries(self):
        '''Test that we don't need to access the objects until we make a change.'''
        self.b.location = self.l1
        self.b.save()
        self.c.location = self.l1
        self.c.save()

        with self.assertNumQueries(1):
            x = SentinelModel.objects.get(pk=self.b.pk)
            x.location = self.l1

        with self.assertNumQueries(1):
            x = SentinelModel.objects.select_related('location__real_parent').get(pk=self.b.pk)
            x.location = self.l2


class TCNoMeta(ClosureModel):
    """A test model without a ClosureMeta."""
    parent = models.ForeignKey(
        "self",
        related_name="children",
        null=True,
        blank=True
    )
    name = models.CharField(max_length=32)

class NoMetaTestCase(TestCase):
    '''Testing models without a ClosureMeta.'''

    def test_basic(self):
        """
        Basic test that you don't need the ClosureMeta class.
        """
        a = TCNoMeta.objects.create(name='a')
        b = TCNoMeta.objects.create(name='b', parent=a)
        c = TCNoMeta.objects.create(name='c', parent=b)
        self.failUnlessEqual(a.get_descendants().count(), 2)
        self.failUnlessEqual(c.get_ancestors().count(), 2)
